import unittest
from pyqint import PyQInt, Molecule
import numpy as np
import multiprocessing
import os

class TestIntegralsOpenMP(unittest.TestCase):

    def test_integrals_h2o_openmp_sto3g(self):
        """
        Test OpenMP integral evaluation for H2O using sto-3g basis set
        """

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule("H2O")
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)
        mol.add_atom('O', 0.0, 0.0, 0.0)
        cgfs, nuclei = mol.build_basis('sto3g')

        # evaluate all integrals
        ncpu = multiprocessing.cpu_count()
        S, T, V, tetensor = integrator.build_integrals_openmp(cgfs, nuclei)

        # load results from npy files
        S_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_overlap_sto3g.npy'))
        T_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_kinetic_sto3g.npy'))
        V_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_nuclear_sto3g.npy'))

        # grab unique two-electron integrals
        teint_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_teint_sto3g.npy'))

        # repackage unique two-electron integrals into a tensor object
        tetensor_result = self.__repackage_tensor(teint_result, len(cgfs))

        # assessment
        self.assertEqual(len(cgfs), 7)
        self.assertEqual(S.shape[0], S.shape[1])
        self.assertEqual(S.shape[0], len(cgfs))
        np.testing.assert_almost_equal(S, S_result, 4)
        np.testing.assert_almost_equal(T, T_result, 4)
        np.testing.assert_almost_equal(V, V_result, 4)
        np.testing.assert_almost_equal(tetensor, tetensor_result, 4)

    def test_integrals_h2o_openmp_sto6g(self):
        """
        Test OpenMP integral evaluation for H2O using sto-6g basis set
        """

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule("H2O")
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)
        mol.add_atom('O', 0.0, 0.0, 0.0)
        cgfs, nuclei = mol.build_basis('sto6g')

        # evaluate all integrals
        ncpu = multiprocessing.cpu_count()
        S, T, V, tetensor = integrator.build_integrals_openmp(cgfs, nuclei)

        # load results from npy files
        S_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_overlap_sto6g.npy'))
        T_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_kinetic_sto6g.npy'))
        V_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_nuclear_sto6g.npy'))
        teint_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_teint_sto6g.npy'))

        # repackage unique two-electron integrals into a tensor object
        tetensor_result = self.__repackage_tensor(teint_result, len(cgfs))

        # assessment
        self.assertEqual(len(cgfs), 7)
        self.assertEqual(S.shape[0], S.shape[1])
        self.assertEqual(S.shape[0], len(cgfs))
        np.testing.assert_almost_equal(S, S_result, 4)
        np.testing.assert_almost_equal(T, T_result, 4)
        np.testing.assert_almost_equal(V, V_result, 4)
        np.testing.assert_almost_equal(tetensor, tetensor_result, 4)

    def test_integrals_h2o_openmp_p321(self):
        """
        Test OpenMP integral evaluation for H2O using p-321 basis set
        """

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule("H2O")
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)
        mol.add_atom('O', 0.0, 0.0, 0.0)
        cgfs, nuclei = mol.build_basis('p321')

        # evaluate all integrals
        ncpu = multiprocessing.cpu_count()
        S, T, V, tetensor = integrator.build_integrals_openmp(cgfs, nuclei)

        # load results from npy files
        S_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_overlap_p321.npy'))
        T_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_kinetic_p321.npy'))
        V_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_nuclear_p321.npy'))
        teint_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_teint_p321.npy'))

        # repackage unique two-electron integrals into a tensor object
        tetensor_result = self.__repackage_tensor(teint_result, len(cgfs))

        # assessment
        self.assertEqual(len(cgfs), 13)
        self.assertEqual(S.shape[0], S.shape[1])
        self.assertEqual(S.shape[0], len(cgfs))
        np.testing.assert_almost_equal(S, S_result, 4)
        np.testing.assert_almost_equal(T, T_result, 4)
        np.testing.assert_almost_equal(V, V_result, 4)
        np.testing.assert_almost_equal(tetensor, tetensor_result, 4)

    def test_integrals_h2o_openmp_p631(self):
        """
        Test OpenMP integral evaluation for H2O using p-631 basis set
        """

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule("H2O")
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)
        mol.add_atom('O', 0.0, 0.0, 0.0)
        cgfs, nuclei = mol.build_basis('p631')

        # evaluate all integrals
        ncpu = multiprocessing.cpu_count()
        S, T, V, tetensor = integrator.build_integrals_openmp(cgfs, nuclei)

        # load results from npy files
        S_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_overlap_p631.npy'))
        T_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_kinetic_p631.npy'))
        V_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_nuclear_p631.npy'))
        teint_result = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_teint_p631.npy'))

        # repackage unique two-electron integrals into a tensor object
        tetensor_result = self.__repackage_tensor(teint_result, len(cgfs))

        # assessment
        self.assertEqual(len(cgfs), 13)
        self.assertEqual(S.shape[0], S.shape[1])
        self.assertEqual(S.shape[0], len(cgfs))
        np.testing.assert_almost_equal(S, S_result, 4)
        np.testing.assert_almost_equal(T, T_result, 4)
        np.testing.assert_almost_equal(V, V_result, 4)
        np.testing.assert_almost_equal(tetensor, tetensor_result, 4)

    def __repackage_tensor(self, teint_result, N):
        # construct integrator object
        integrator = PyQInt()

        # repackage unique two-electron integrals into a tensor object
        tetensor_result = np.zeros((N,N,N,N))
        for i in range(N):
            for j in range(N):
                for k in range(N):
                    for l in range(N):
                        idx = integrator.teindex(i,j,k,l)
                        tetensor_result[i,j,k,l] = teint_result[idx]

        return tetensor_result

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import PyQInt, GTO, Molecule
import numpy as np

class TestNuclear(unittest.TestCase):

    def test_gto_nuclear(self):
        """
        Test nuclear attraction integral for GTOs

        V^{(c)}_ij = <gto_i | -1 / |r-Rc| | gto_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # test GTO
        gto1 = GTO(0.154329, [0.0, 0.0, 0.0], 3.425251, 0, 0, 0)
        gto2 = GTO(0.535328, [0.0, 0.0, 0.0], 0.623914, 0, 0, 0)
        gto3 = GTO(0.444635, [0.0, 0.0, 0.0], 0.168855, 0, 0, 0)
        nuclear = integrator.nuclear_gto(gto1, gto1, [0.0, 0.0, 1.0])
        result = -0.31049036979675293
        np.testing.assert_almost_equal(nuclear, result, 4)

    def test_cgf_nuclear(self):
        """
        Test nuclear attraction integrals for contracted Gaussians

        V^{(c)}_ij = <cgf_i | -Zc / |r-Rc| | cgf_j>
        """

        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.0, 0.0, 1.4)
        cgfs, nuclei = mol.build_basis('sto3g')

        V1 = np.zeros((2,2))
        V1[0,0] = integrator.nuclear(cgfs[0], cgfs[0], cgfs[0].p, 1)
        V1[0,1] = V1[1,0] = integrator.nuclear(cgfs[0], cgfs[1], cgfs[0].p, 1)
        V1[1,1] = integrator.nuclear(cgfs[1], cgfs[1], cgfs[0].p, 1)

        V2 = np.zeros((2,2))
        V2[0,0] = integrator.nuclear(cgfs[0], cgfs[0], cgfs[1].p, 1)
        V2[0,1] = V2[1,0] = integrator.nuclear(cgfs[0], cgfs[1], cgfs[1].p, 1)
        V2[1,1] = integrator.nuclear(cgfs[1], cgfs[1], cgfs[1].p, 1)

        V11 = -1.2266135215759277
        V12 = -0.5974172949790955
        V22 = -0.6538270711898804
        np.testing.assert_almost_equal(V1[0,0], V11, 4)
        np.testing.assert_almost_equal(V1[1,1], V22, 4)
        np.testing.assert_almost_equal(V1[0,1], V12, 4)
        np.testing.assert_almost_equal(V2[0,0], V22, 4)
        np.testing.assert_almost_equal(V2[1,1], V11, 4)
        np.testing.assert_almost_equal(V2[0,1], V12, 4)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import Molecule, HF
import numpy as np

class TestHF(unittest.TestCase):

    def testHartreeFockWater(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        mol.add_atom('O', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)

        results = perform_hf(mol)

        # check that energy matches
        np.testing.assert_almost_equal(results['energy'], -73.21447132, 4)

def perform_hf(mol):
    results = HF().rhf(mol, 'sto3g', tolerance=1e-9)
    return results

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .pyqint import PyQInt
import numpy as np
import random
import scipy.optimize
from scipy.stats import ortho_group

class FosterBoys:
    """
    Routine for constructing localized orbitals using Foster-Boys procedure
    """
    def __init__(self, res, seed=None):
        # copy objects from Hartree-Fock result dictionary
        self.orbc_canonical = res['orbc']
        self.orbe_canonical = res['orbe']
        self.mol = res['mol']
        self.nelec = res['nelec']
        self.H = res['fock']
        self.cgfs = res['cgfs']
        self.maxiter = 1000
        self.occ = [1 if i < self.nelec//2 else 0 for i in range(0, len(self.cgfs))]
        self.rngseed = seed
        self.rng = np.random.default_rng(seed=self.rngseed)
        
        # construct dipole tensor
        self.dipol = self.__build_dipole_tensor(self.cgfs)
    
    def run(self, nr_runners=1):
        """
        Perform the Foster-Boys localization procedure

        Because the Foster-Boys procedure uses random initialization, it is possible that
        the algorithm ends up in a local minima. To circumvent this, the user can use
        a number of runners and automatically use the best result found.
        """
        bestres = None
        bestr2 = 0.0

        for i in range(0, nr_runners):
            res = self.__single_runner()
            if res['r2final'] > bestr2:
                bestres = res

        return bestres

    def __single_runner(self):
        """
        Perform the Foster-Boys optimization routine
        """
        # start with a random unitary transformation
        C = self.__construct_random_orthogonal_transformation(self.orbc_canonical,
                                                              self.nelec)

        old_r2 = 0.0 # start assuming orbital centroids lie at origin
        diff = 1
        nriter = 0
        while diff > 1e-7 and nriter < self.maxiter:
            C, r2 = self.__perform_foster_boys_mixing_scipy(C)
            diff = np.abs(r2 - old_r2)
            old_r2 = r2
            nriter += 1
        
        if nriter == self.maxiter:
            raise Exception('Foster-Boys procedure did not converge.')

        orbe, orbc = self.__calculate_molecular_orbital_energies(C)

        result = {
            'orbe': orbe,
            'orbc': orbc,
            'nriter': nriter,
            'mol': self.mol,
            'r2start': self.__calculate_r2(self.orbc_canonical),
            'r2final': self.__calculate_r2(orbc),
            'nelec': self.nelec,
            'cgfs': self.cgfs,
        }

        return result
    
    def __perform_foster_boys_mixing_scipy(self, C):
        """
        Successively perform n(n-1)/2 optimization among the occupied
        molecular orbitals using the scipy minimize function
        """
        r2start = self.__calculate_r2(C)
        r2final = r2start
        
        N = self.nelec // 2
        for i in range(N):
            for j in range(i+1, N):
                
                # optimize local pair of MOs on the interval -(pi,pi)
                res = scipy.optimize.minimize(self.__evaluate_orbital_centroids, 
                                              0.0,
                                              args=(C, i, j),
                                              bounds=[(-np.pi, np.pi)],
                                              tol=1e-12)
                
                alpha = res.x[0]
                Cnew = C.copy()
                Cnew[:,i] = np.cos(alpha) * C[:,i] + np.sin(alpha) * C[:,j]
                Cnew[:,j] = -np.sin(alpha) * C[:,i] + np.cos(alpha) * C[:,j]
                
                r2 = self.__calculate_r2(Cnew)
                if(r2 > r2start):
                    C = Cnew.copy()
                    r2final = r2start
        
        return C, r2final
    
    def __evaluate_orbital_centroids(self, alpha, C, i, j):
        """
        Evaluate the result of r2 after a 2x2 rotation
        between solutions i and j
        """
        
        Cnew = C.copy()
        Cnew[:,i] = np.cos(alpha) * C[:,i] + np.sin(alpha) * C[:,j]
        Cnew[:,j] = -np.sin(alpha) * C[:,i] + np.cos(alpha) * C[:,j]
        r2 = self.__calculate_r2(Cnew)
        
        return -r2
    
    def __calculate_r2(self, C):
        """
        Calculate r2 from coefficient matrix
        """
        # use efficient Einstein summation
        # i -> molecular orbitals
        # j -> basis functions 1
        # k -> basis functions 2
        # l -> cartesian directions dipole moment
        dipolest = np.einsum('ji,ki,jkl->il', C, C, self.dipol)
        r2 = np.einsum('ij,i->', dipolest**2, self.occ)
        
        return r2
    
    def __construct_random_orthogonal_transformation(self, C, nelec, nops=100):
        """
        Construct unitary transformation matrix via a series of two-dimensional
        unitary transformations among the occupied molecular orbitals
        """
        for i in range(0,nops):
            Cnew = C.copy()

            # randomly pick two numbers among the occupied orbitals
            n = self.rng.choice(range(nelec//2), size=2, replace=False)
            
            # and mix them by a random angle
            gamma = self.rng.uniform() * 2.0 * np.pi
            
            for j in range(0,len(C)):
                Cnew[j,n[0]] = np.cos(gamma) * C[j,n[0]] + np.sin(gamma) * C[j,n[1]]
                Cnew[j,n[1]] = -np.sin(gamma) * C[j,n[0]] + np.cos(gamma) * C[j,n[1]]

            C = Cnew.copy()

        return Cnew

    def __construct_random_orthogonal_transformation_from_orthogroup(self, C, nelec, nops=100):
        """
        Create orthogonal transformation usign the scipy ortho_group function

        This routine should in theory act as a substitute for the
        __construct_random_orthogonal_transformation() routine, but it is not working
        appropriately. Leaving it here for potential future development.
        """
        N = nelec//2
        orthomat = ortho_group.rvs(N)
        identitymat = np.identity(len(C) - N)
        T = scipy.linalg.block_diag(orthomat, identitymat)

        return T @ C @ T.transpose()
    
    def __build_dipole_tensor(self, cgfs):
        """
        Build a dipole tensor
        """
        # create new cgfs
        N = len(cgfs)
        mat = np.zeros((N,N,3))
        integrator = PyQInt()
        for i,cgf1 in enumerate(cgfs):
            for j,cgf2 in enumerate(cgfs):
                for k in range(0,3):
                    mat[i,j,k] = integrator.dipole(cgf1, cgf2, k, 0.0)
        
        return mat

    def __calculate_molecular_orbital_energies(self, C):
        """
        Calculate the one-electron MO energies from the Hamiltonian matrix
        and the coefficient matrix, both in their original basis

        Return *ordered* list of eigenvalue and -vector pairs
        """
        orbe = np.zeros(len(C))
        for i in range(len(C)):
            orbe[i] = C[:,i].dot(self.H.dot(C[:,i]))

        # produce list of indices for eigenvalues in ascending order
        oidx = np.argsort(orbe)

        return orbe[oidx], C[:,oidx]



import unittest
from pyqint import MoleculeBuilder
import numpy as np
import os

class TestMoleculeBuilder(unittest.TestCase):
    
    def test_construction(self):
        """
        Build a molecule from point group
        """
        mol = MoleculeBuilder().build_complex_td(1.0, 'C', 'H')

    def test_load_molecule_from_name(self):
        """
        Build a molecule from filename
        """
        mol = MoleculeBuilder().from_name('ch4')

        np.testing.assert_almost_equal(mol.get_atoms()[0][1], 
                                       np.array([0.0,0.0,0.0], dtype=np.float64))
        np.testing.assert_almost_equal(mol.get_atoms()[1][1], 
                                       np.array([0.6327670,0.6327670,0.6327670]) * 1.8897259886)
        np.testing.assert_equal(mol.get_atoms()[0][0], 'C')
        
        mol.build_basis('sto3g')
        np.testing.assert_equal(mol.get_nelec(), 10)
        
    def test_load_molecule_from_file(self):
        """
        Build a molecule from file
        """
        fname = os.path.join(os.path.dirname(__file__), 'results', 'ch4.xyz')
        mol = MoleculeBuilder().from_file(fname)

        np.testing.assert_almost_equal(mol.get_atoms()[0][1], 
                                       np.array([0.0,0.0,0.0], dtype=np.float64))
        np.testing.assert_almost_equal(mol.get_atoms()[1][1], 
                                       np.array([0.6327670,0.6327670,0.6327670]) * 1.8897259886)
        np.testing.assert_equal(mol.get_atoms()[0][0], 'C')
        
        mol.build_basis('sto3g')
        np.testing.assert_equal(mol.get_nelec(), 10)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import PyQInt, Molecule
from copy import deepcopy
import numpy as np
import os

class TestNuclearDeriv(unittest.TestCase):

    def testNuclearRepulsionDerivatives(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        mol.add_atom('O', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)

        cgfs, nuclei = mol.build_basis('sto3g')

        forces_fd = calculate_deriv_nuclear_repulsion_finite_difference(mol)

        forces = np.zeros(forces_fd.shape)
        for i in range(0, len(mol.get_atoms())): # loop over nuclei
            for j in range(0, 3): # loop over directions
                forces[i,j] = deriv_nuclear_repulsion(nuclei, i, j)

        np.testing.assert_almost_equal(forces, forces_fd, 4)

    def test_derivatives_h2o_fulltest(self):
        """
        Test Derivatives of water
        """
        # build integrator object
        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule('H2O')
        mol.add_atom('O', 0.00000, -0.07579, 0.00000, unit='angstrom')
        mol.add_atom('H', 0.86681, 0.60144, 0.00000, unit='angstrom')
        mol.add_atom('H',  -0.86681, 0.60144, 0.00000, unit='angstrom')
        cgfs, nuclei = mol.build_basis('sto3g')

        # get position and charge
        probe_atom = 0
        O = nuclei[probe_atom][0]
        Ochg = nuclei[probe_atom][1]

        # load results from file
        fname = os.path.join(os.path.dirname(__file__), 'results', 'nuclear_deriv_h2o.txt')
        vals = np.loadtxt(fname).reshape((len(cgfs), len(cgfs), 3, 3))
        for i in range(0, len(cgfs)): # loop over cgfs
            for j in range(0, len(cgfs)): # loop over cgfs
                for k in range(0,3):  # loop over nuclei
                    for l in range(0,3):  # loop over directions
                        force = integrator.nuclear_deriv(cgfs[i], cgfs[j], O, Ochg, nuclei[k][0], l)
                        np.testing.assert_almost_equal(force, vals[i,j,k,l], 4)

def calculate_force_finite_difference(cgf_id1, cgf_id2, nuc_id, coord, probe_atom):
    # build integrator object
    integrator = PyQInt()

    # distance
    diff = 0.00001

    vals = np.zeros(2)
    for i,v in enumerate([-1,1]):
        # build hydrogen molecule
        mol = Molecule('H2O')
        mol.add_atom('O', 0.00000, -0.07579, 0.00000, unit='angstrom')
        mol.add_atom('H', 0.86681, 0.60144, 0.00000, unit='angstrom')
        mol.add_atom('H',  -0.86681, 0.60144, 0.00000, unit='angstrom')

        # adjust molecule
        mol.get_atoms()[nuc_id][1][coord] += v * diff / 2

        # build basis
        cgfs, nuclei = mol.build_basis('sto3g')

        # calculate values
        O = nuclei[probe_atom][0]
        Ochg = nuclei[probe_atom][1]
        vals[i] = integrator.nuclear(cgfs[cgf_id1], cgfs[cgf_id2], O, Ochg)

    return (vals[1] - vals[0]) / diff

def energy_nuclear_repulsion(nuclei):
    energy = 0.0
    for i in range(0, len(nuclei)):
        for j in range(i+1, len(nuclei)):
            r = np.linalg.norm(np.array(nuclei[i][0]) - np.array(nuclei[j][0]))
            energy += nuclei[i][1] * nuclei[j][1] / r
    return energy

def deriv_nuclear_repulsion(nuclei, nucid, coord):
    Vnn = 0.0
    pc = nuclei[nucid][0]
    for i in range(0, len(nuclei)):
        if nucid != i:
            pi = nuclei[i][0]
            Vnn += nuclei[nucid][1] * nuclei[i][1] * (pi[coord] - pc[coord]) / np.linalg.norm(pi - pc)**3

    return Vnn

def calculate_deriv_nuclear_repulsion_finite_difference(mol):
    forces = np.zeros((3,3))

    sz = 0.0001

    for i in range(0, len(mol.get_atoms())): # loop over nuclei
        for j in range(0, 3): # loop over directions
            mol1 = deepcopy(mol)
            mol1.get_atoms()[i][1][j] -= sz / 2
            mol2 = deepcopy(mol)
            mol2.get_atoms()[i][1][j] += sz / 2

            cgfs, nuclei1= mol1.build_basis('sto3g')
            cgfs, nuclei2 = mol2.build_basis('sto3g')

            energy1 = energy_nuclear_repulsion(nuclei1)
            energy2 = energy_nuclear_repulsion(nuclei2)

            forces[i,j] = (energy2 - energy1) / sz

    return forces

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .pyqint import PyQInt
import numpy as np

class COHP:
    """
    Class for performing Crystal Orbital Hamilton Population Analysis
    """
    def __init__(self, res):
        # copy objects from Hartree-Fock result dictionary
        self.orbc_canonical = res['orbc']
        self.orbe_canonical = res['orbe']
        self.nelec = res['nelec']
        self.nuclei = res['nuclei']
        self.H = res['fock']
        self.cgfs = res['cgfs']
        self.maxiter = 1000
        self.occ = [1 if i < self.nelec//2 else 0 for i in range(0, len(self.cgfs))]

    def run(self, C, n1, n2):
        """
        Run the COHP algorithm, take a coefficient matrix as input

        C: coefficient matrix
        n1: index of nucleus 1
        n2: index of nucleus 2
        """
        if n1 == n2:
            raise Exception('Cannot perform COHP for the same atoms')

        # figure out which basis functions below to which nucleus
        cgfs1 = []
        cgfs2 = []
        nuc1 = self.nuclei[n1][0]
        nuc2 = self.nuclei[n2][0]
        for i,cgf in enumerate(self.cgfs):
            if np.linalg.norm(cgf.p - nuc1) < 1e-3:
                cgfs1.append(i)
            if np.linalg.norm(cgf.p - nuc2) < 1e-3:
                cgfs2.append(i)

        N = len(C)
        cohp = np.zeros(N)
        for k in range(0, N): # loop over molecular orbitals
            for i in cgfs1: # loop over atom A
                for j in cgfs2: # loop over atom B
                    cohp[k] += 2.0 * self.H[i,j] * C[i,k] * C[j,k]

        return cohp



import unittest
from pyqint import PyQInt, Molecule
import numpy as np

class TestSafetyGuards(unittest.TestCase):
    """
    Tests certain safety features preventing user errors
    """

    def test_safety_plot_wavefunction(self):
        """
        Test for correct array sizes
        """
        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.00, -0.7)  # distances in Bohr lengths
        mol.add_atom('H', 0.0, 0.00, 0.7)   # distances in Bohr lengths
        cgfs, nuclei = mol.build_basis('sto3g')
        
        # construct integrator object
        integrator = PyQInt()
        
        # build grid points
        x = np.linspace(-2, 2, 6, endpoint=True)
        grid = np.flipud(np.vstack(np.meshgrid(x, x, x, indexing='ij')).reshape(3,-1)).T
        
        # put in matrix object -> should yield error
        c = np.identity(2)
        with self.assertRaises(TypeError) as raises_cm:
            integrator.plot_wavefunction(grid, c, cgfs)

        exception = raises_cm.exception
        self.assertTrue('arrays can be converted to Python scalars' in exception.args[0])
        
        # put in matrix object -> should yield error
        c = np.identity(3)
        with self.assertRaises(Exception) as raises_cm:
            integrator.plot_wavefunction(grid, c, cgfs)

        exception = raises_cm.exception
        self.assertEqual(exception.args, ('Dimensions of cgf list and coefficient matrix do not match (2 != 3)',))
        
        # put in matrix object -> should yield error
        c = np.ones(3)
        with self.assertRaises(Exception) as raises_cm:
            integrator.plot_wavefunction(grid, c, cgfs)
            
        exception = raises_cm.exception
        self.assertEqual(exception.args, ('Dimensions of cgf list and coefficient matrix do not match (2 != 3)',))
        
    def test_safety_plot_gradient(self):
        """
        Test for correct array sizes
        """
        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.00, -0.7)  # distances in Bohr lengths
        mol.add_atom('H', 0.0, 0.00, 0.7)   # distances in Bohr lengths
        cgfs, nuclei = mol.build_basis('sto3g')
        
        # construct integrator object
        integrator = PyQInt()
        
        # build grid points
        x = np.linspace(-2, 2, 6, endpoint=True)
        grid = np.flipud(np.vstack(np.meshgrid(x, x, x, indexing='ij')).reshape(3,-1)).T
        
        # put in matrix object -> should yield error
        c = np.identity(2)
        with self.assertRaises(TypeError) as raises_cm:
            integrator.plot_gradient(grid, c, cgfs)

        exception = raises_cm.exception
        self.assertEqual(type(exception), TypeError)
        
        # put in matrix object -> should yield error
        c = np.identity(3)
        with self.assertRaises(Exception) as raises_cm:
            integrator.plot_gradient(grid, c, cgfs)

        exception = raises_cm.exception
        self.assertEqual(exception.args, ('Dimensions of cgf list and coefficient matrix do not match (2 != 3)',))
        
        # put in matrix object -> should yield error
        c = np.ones(3)
        with self.assertRaises(Exception) as raises_cm:
            integrator.plot_gradient(grid, c, cgfs)
            
        exception = raises_cm.exception
        self.assertEqual(exception.args, ('Dimensions of cgf list and coefficient matrix do not match (2 != 3)',))

if __name__ == '__main__':
    unittest.main()



import bpy
import numpy as np
import os
import time
import json

#
# IMPORTANT
#
# Do not run this script natively. This script is meant to be run in Blender
# via one of the call routines
#

with open(os.path.join(os.path.dirname(__file__), 'manifest.json')) as f:
    manifest = json.load(f)

def main():
    # set the scene
    settings = {
        'resolution': 512,
        'camera_location': (-10,0,0),
        'camera_rotation': (np.pi/2,0,-np.pi/2),
        'camera_scale' : 10
    }
    set_environment(settings)

    # read molecule file and load it
    mol = read_xyz(manifest['xyzfile'])
    create_atoms(mol)
    create_bonds(mol)

    add_isosurface(manifest['mo_name'],
                   manifest['mo_neg_path'],
                   manifest['mo_pos_path'])

    render_scene(manifest['png_output'])

def add_isosurface(label, filename_neg, filename_pos):
    """
    Add the two isosurfaces
    """
    bpy.ops.import_mesh.ply(
        filepath=filename_neg
    )
    obj = bpy.context.object
    bpy.ops.object.shade_smooth()
    obj.data.materials.append(create_material('matneg', manifest['mo_colors']['neg'], alpha=0.5))
    obj.name = 'isosurface ' + label + '_neg'

    obj = bpy.ops.import_mesh.ply(
        filepath=filename_pos
    )
    bpy.ops.object.shade_smooth()
    obj = bpy.context.object
    obj.data.materials.append(create_material('matpos', manifest['mo_colors']['pos'], alpha=0.5))
    obj.name = 'isosurface ' + label + '_pos'

def create_atoms(mol):
    """
    Create atoms
    """
    for i,at in enumerate(mol):
        scale = manifest['atom_radii'][at[0]]
        bpy.ops.surface.primitive_nurbs_surface_sphere_add(
            radius=scale,
            enter_editmode=False,
            align='WORLD',
            location=at[1])
        obj = bpy.context.view_layer.objects.active
        obj.name = "atom-%s-%03i" % (at[0],i)
        bpy.ops.object.shade_smooth()

        # set a material
        mat = create_material(at[0], manifest['atom_colors'][at[0]])
        print(mat)
        obj.data.materials.append(mat)

def create_bonds(mol):
    """
    Create bonds between atoms
    """
    # set default orientation of bonds (fixed!)
    z = np.array([0,0,1])

    # add new bonds material if it does not yet exist
    matbond = create_material('bond', '222222')

    for i,at1 in enumerate(mol):
        r1 = np.array(at1[1])
        for j,at2 in enumerate(mol[i+1:]):
            r2 = np.array(at2[1])
            dist = np.linalg.norm(r2 - r1)

            # only create a bond if the distance is less than 1.5 A
            if dist < 2.5:
                axis = np.cross(z,r2-r1)
                if np.linalg.norm(axis) < 1e-5:
                    axis = np.array([0,0,1])
                    angle = 0.0
                else:
                    axis /= np.linalg.norm(axis)
                    angle = np.arccos(np.dot(r2-r1,z)/dist)

                bpy.ops.surface.primitive_nurbs_surface_cylinder_add(
                    enter_editmode=False,
                    align='WORLD',
                    location=tuple((r1 + r2) / 2)
                )

                obj = bpy.context.view_layer.objects.active
                obj.scale = (manifest['bond_thickness'], manifest['bond_thickness'], dist/2)
                obj.rotation_mode = 'AXIS_ANGLE'
                obj.rotation_axis_angle = (angle, axis[0], axis[1], axis[2])

                obj.name = "bond-%s-%03i-%s-%03i" % (at1[0],i,at2[0],j)
                bpy.ops.object.shade_smooth()
                obj.data.materials.append(matbond)

def set_environment(settings):
    """
    Specify canvas size, remove default objects, reset positions of
    camera and light, define film and set background
    """
    print('Set render engine to: CYCLES')
    bpy.context.scene.render.engine = 'CYCLES'
    print('Set rendering device to GPU')
    bpy.context.scene.cycles.device = 'GPU'
    bpy.context.scene.render.resolution_x = settings['resolution']
    bpy.context.scene.render.resolution_y = settings['resolution']
    print('Setting resolution to: ', settings['resolution'])
    bpy.context.scene.cycles.samples = 1024
    bpy.context.scene.cycles.tile_size = 2048

    # remove cube
    if 'Cube' in bpy.data.objects:
        o = bpy.data.objects['Cube']
        bpy.data.objects.remove(o, do_unlink=True)

    # set camera into default position
    bpy.data.objects['Camera'].location = tuple(settings['camera_location'])
    bpy.data.objects['Camera'].rotation_euler = tuple(settings['camera_rotation'])
    bpy.data.objects['Camera'].data.clip_end = 1000
    bpy.data.objects['Camera'].data.type = 'ORTHO'
    bpy.data.objects['Camera'].data.ortho_scale = settings['camera_scale']

    # set lights
    bpy.data.objects['Light'].data.type = 'AREA'
    bpy.data.objects['Light'].data.energy = 1e4
    bpy.data.objects['Light'].location = (-10,10,10)
    bpy.data.objects['Light'].rotation_euler = tuple(np.radians([55, 0, 225]))
    bpy.data.objects['Light'].data.shape = 'DISK'
    bpy.data.objects['Light'].data.size = 10

    # set film
    bpy.context.scene.render.film_transparent = True

    # set background
    bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (1,1,1,1)
    bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[1].default_value = 1

def create_material(name, color, alpha=1.0):
    """
    Build a new material
    """
    # early exit if material already exists
    if name in bpy.data.materials:
        return bpy.data.materials[name]

    mat = bpy.data.materials.new(name)
    mat.use_nodes = True

    # set base color
    mat.node_tree.nodes["Principled BSDF"].inputs[0].default_value = hex2rgbtuple(color)

    # subsurface modifier
    mat.node_tree.nodes["Principled BSDF"].inputs[1].default_value = 0.2

    # set subsurface radii
    mat.node_tree.nodes["Principled BSDF"].inputs[2].default_value = (0.3,0.3,0.3)

    # set subsurface color
    mat.node_tree.nodes["Principled BSDF"].inputs[3].default_value = hex2rgbtuple('000000')

    # metallic
    mat.node_tree.nodes["Principled BSDF"].inputs[4].default_value = 0.3

    # roughness
    mat.node_tree.nodes["Principled BSDF"].inputs[7].default_value = 0.05

    # alpha
    mat.node_tree.nodes["Principled BSDF"].inputs[21].default_value = alpha

    return mat

def render_scene(outputfile, samples=512):
    """
    Render the scene
    """
    bpy.context.scene.cycles.samples = samples

    print('Start render')
    start = time.time()
    bpy.data.scenes['Scene'].render.filepath = outputfile
    bpy.ops.render.render(write_still=True)
    end = time.time()
    print('Finished rendering frame in %.1f seconds' % (end - start))

def read_xyz(filename):
    f = open(filename)
    nratoms = int(f.readline())
    f.readline() # skip line

    angtobohr = 1.8897259886

    mol = []
    for i in range(0,nratoms):
        pieces = f.readline().split()
        mol.append(
            [pieces[0],
            (
                float(pieces[1]) * angtobohr,
                float(pieces[2]) * angtobohr,
                float(pieces[3]) * angtobohr
            )]
        )

    return mol

def hex2rgbtuple(hexcode):
    """
    Convert 6-digit color hexcode to a tuple of floats
    """
    hexcode += "FF"
    hextuple = tuple([int(hexcode[i:i+2], 16)/255.0 for i in [0,2,4,6]])

    return tuple([color_srgb_to_scene_linear(c) for c in hextuple])

def color_srgb_to_scene_linear(c):
    """
    Convert RGB to sRGB
    """
    if c < 0.04045:
        return 0.0 if c < 0.0 else c * (1.0 / 12.92)
    else:
        return ((c + 0.055) * (1.0 / 1.055)) ** 2.4

if __name__ == '__main__':
    main()



import unittest
from pyqint import Molecule, HF
import numpy as np

class TestMoleculeOrder(unittest.TestCase):
    
    def testOrder(self):
        """
        Test whether the same calculation but with a different order in the
        atoms gives the same result. Note that this test does not pass when
        the energetic convergence criterium is set too loose (e.g. 1e-5)
        """
        results1 = build_mol_order1()
        results2 = build_mol_order2()
   
        en1 = results1['energy']
        en2 = results2['energy']
        
        np.testing.assert_almost_equal(en1, en2)
    
def build_mol_order1():
    """
    Test Hartree-Fock calculation for Ethylene
    
    Data is compared to results obtained from Gaussian
    """
    mol = Molecule()
    mol.add_atom('C', -0.6530176758,  0.0000000000 ,0.0000000000, unit='angstrom')
    mol.add_atom('C',  0.6530176758,  0.0000000000 ,0.0000000000, unit='angstrom')
    mol.add_atom('H', -1.2288875372, -0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('H', -1.2288875372,  0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('H',  1.2288875372,  0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('H',  1.2288875372, -0.9156191261 ,0.0000000000, unit='angstrom')

    results = HF().rhf(mol, basis='sto3g', tolerance=1e-12)
    
    return results
    
def build_mol_order2():
    """
    Test Hartree-Fock calculation for Ethylene
    
    Data is compared to results obtained from Gaussian
    """
    mol = Molecule()
    mol.add_atom('C', -0.6530176758,  0.0000000000 ,0.0000000000, unit='angstrom')
    mol.add_atom('H', -1.2288875372, -0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('H', -1.2288875372,  0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('C',  0.6530176758,  0.0000000000 ,0.0000000000, unit='angstrom')
    mol.add_atom('H',  1.2288875372,  0.9156191261 ,0.0000000000, unit='angstrom')
    mol.add_atom('H',  1.2288875372, -0.9156191261 ,0.0000000000, unit='angstrom')

    results = HF().rhf(mol, basis='sto3g', tolerance=1e-12)
    
    return results

if __name__ == '__main__':
    unittest.main()



import unittest
import pyqint
from pyqint import PyQInt, Molecule

class TestString(unittest.TestCase):

    def test_strings(self):
        """
        Test automatic integral evaluation for hydrogen molecule
        """

        # construct integrator object
        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule('H2')
        mol.add_atom('H', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.0, 0.0, 1.4)
        cgfs, nuclei = mol.build_basis('sto3g')

        ans = "CGF; R=(0.000000,0.000000,0.000000)\n"
        ans += " 01 | GTO : c=0.154329, alpha=3.425251, l=0, m=0, n=0, R=(0.000000,0.000000,0.000000)\n"
        ans += " 02 | GTO : c=0.535328, alpha=0.623914, l=0, m=0, n=0, R=(0.000000,0.000000,0.000000)\n"
        ans += " 03 | GTO : c=0.444635, alpha=0.168855, l=0, m=0, n=0, R=(0.000000,0.000000,0.000000)\n"
        self.assertEqual(str(cgfs[0]), ans)

        ans = "Molecule: H2\n"
        ans += " H (0.000000,0.000000,0.000000)\n"
        ans += " H (0.000000,0.000000,1.400000)\n"
        self.assertEqual(str(mol), ans)

    def test_version(self):
        strver = pyqint.__version__.split('.')
        self.assertTrue(strver[0].isnumeric())
        self.assertTrue(strver[1].isnumeric())
        self.assertTrue(strver[2].isnumeric())

if __name__ == '__main__':
    unittest.main()



from .pyqint import PyGTO

class GTO:
    """
    Primitive Gaussian Type Orbital
    """
    def __init__(self, _c, _p, _alpha, _l, _m, _n):
        """
        Default constructor
        """
        self.c = _c
        self.p = _p
        self.alpha = _alpha
        self.l = _l
        self.m = _m
        self.n = _n

        self.gto = PyGTO(self.c, self.p, self.alpha, self.l, self.m, self.n)

    def __setstate__(self, d):
        """
        Set the state using tuple d

        Rebuilds the PyGTO C++ class
        """
        self.c = d[0]
        self.p = d[1]
        self.alpha = d[2]
        self.l = d[3]
        self.m = d[4]
        self.n = d[5]
        self.gto = PyGTO(self.c, self.p, self.alpha, self.l, self.m, self.n)

    def __str__(self):
        return "GTO : c=%f, alpha=%f, l=%i, m=%i, n=%i, R=(%f,%f,%f)\n" \
            % (self.c, self.alpha, self.l, self.m, self.n, self.p[0], self.p[1], self.p[2])

    def __reduce__(self):
        """
        Used to pickle the class
        """
        return (self.__class__, (self.c, self.p, self.alpha, self.l, self.m, self.n))

    def get_amp(self, x, y, z):
        return self.gto.get_amp(x, y, z)

    def get_norm(self):
        """
        Returns the normalization constant for the GTO
        """
        return self.gto.get_norm()


from .gto import GTO
from .pyqint import PyCGF
from . import spherical_harmonics as sh

class CGF:
    """
    Contracted Gaussian Type Orbital
    """
    def __init__(self, _p=[0,0,0]):
        """
        Default constructor
        """
        self.gtos = []
        self.p = _p
        self.cgf = PyCGF(self.p)

    def __setstate__(self, d):
        """
        Set the state using tuple d

        Rebuilds the PyCGF C++ class and
        adds gto object to the class
        """
        self.p = d[0]
        self.gtos = d[1]
        self.cgf = PyCGF(self.p)
        for gto in self.gtos:
            self.cgf.add_gto(gto.c, gto.alpha, gto.l, gto.m, gto.n)

    def __reduce__(self):
        """
        Used to pickle the class
        """
        return (self.__class__, tuple([self.p]), (self.p, self.gtos))

    def __str__(self):
        """
        Get string representation of the Contracted Gaussian Functional
        """
        res = "CGF; R=(%f,%f,%f)\n" % tuple(self.p)
        for i,gto in enumerate(self.gtos):
            res += " %02i | %s" % (i+1, str(gto))
        return res

    def add_gto(self, c, alpha, l, m, n):
        """
        Add Gaussian Type Orbital to Contracted Gaussian Function
        """
        self.gtos.append(GTO(c, self.p, alpha, l, m, n))
        self.cgf.add_gto(c, alpha, l, m, n)

    def add_spherical_gto(self, c, alpha, l, m):
        """
        Add Spherical Gaussian Type Orbital to Contracted Gaussian Function.
        l and m are the coefficients of the requested spherical harmonic function. 
        l must be <= 6 and -l <= m <= l.
        """
        if not l <= 6 or not abs(m) <=l:
            raise ValueError("l must be <= 6 and -l <= m <= l")
        for gto in sh.spherical_harmonics[l][m]:
            self.add_gto(gto[0] * c, alpha, gto[1][0], gto[1][1], gto[1][2])

    def get_amp_f(self, x, y, z):
        """
        Get the amplitude of the wave function at position r
        """
        return self.cgf.get_amp_f(x, y, z)

    def get_amp(self, r):
        """
        Get the amplitude of the wave function at position r
        """
        return self.cgf.get_amp_f(r[0], r[1], r[2])

    def get_grad_f(self, x, y, z):
        """
        Get the gradient (3-vector) of the wave function at position r
        """
        return self.cgf.get_grad_f(x, y, z)

    def get_grad(self, r):
        """
        Get the gradient (3-vector) of the wave function at position r
        """
        return self.cgf.get_grad_f(r[0], r[1], r[2])



import unittest
from pyqint import PyQInt, Molecule
import numpy as np
import os

class TestPlotData(unittest.TestCase):

    def test_plot_wavefunction(self):
        """
        Test generation of wave function
        """
        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.00, -0.7)  # distances in Bohr lengths
        mol.add_atom('H', 0.0, 0.00, 0.7)   # distances in Bohr lengths
        cgfs, nuclei = mol.build_basis('sto3g')
        
        # construct integrator object
        integrator = PyQInt()
        
        # build grid points
        x = np.linspace(-2, 2, 6, endpoint=True)
        coord = np.flipud(np.vstack(np.meshgrid(x, x, x, indexing='ij')).reshape(3,-1)).T
        
        c = np.array([0.54893397, 0.54893397])
        wf = integrator.plot_wavefunction(coord, c, cgfs)
        
        res = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2_wf.npy'))
        np.testing.assert_almost_equal(wf, res, 4)
        
    def test_plot_gradient(self):
        """
        Test generation of wave function
        """
        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.00, -0.7)  # distances in Bohr lengths
        mol.add_atom('H', 0.0, 0.00, 0.7)   # distances in Bohr lengths
        cgfs, nuclei = mol.build_basis('sto3g')
        
        # construct integrator object
        integrator = PyQInt()
                       
        c = np.array([0.54893397, 0.54893397])
        
        # test couple of single points
        sc = np.array([[-2.,-2.,-2.]])
        grad = integrator.plot_gradient(sc, c, cgfs)
        self.assertEqual(grad.shape[0], 1)
        self.assertEqual(grad.shape[1], 3)
        res = np.array([[0.00926244,0.00926244,0.0076777]])
        np.testing.assert_almost_equal(grad, res, 4)
        
        sc = np.array([[-2., 2.,-2.]])
        grad = integrator.plot_gradient(sc, c, cgfs)
        self.assertEqual(grad.shape[0], 1)
        self.assertEqual(grad.shape[1], 3)
        res = np.array([[0.00926244, -0.00926244, 0.0076777]])
        np.testing.assert_almost_equal(grad, res, 4)
        
        # test two points
        tc = np.array([[-2.,-2.,-2.],[-2., 2.,-2.]])
        grad = integrator.plot_gradient(tc, c, cgfs)
        self.assertEqual(grad.shape[0], 2)
        self.assertEqual(grad.shape[1], 3)
        res = np.array([[0.00926244,0.00926244,0.0076777],[0.00926244, -0.00926244, 0.0076777]])
        np.testing.assert_almost_equal(grad, res, 4)
        
        # test grid of points
        # x = np.linspace(-2, 2, 6, endpoint=True)
        # coord = np.flipud(np.vstack(np.meshgrid(x, x, x, indexing='ij')).reshape(3,-1)).T
        coord = integrator.build_rectgrid3d(-2, 2, 6, endpoint=True)
        grad = integrator.plot_gradient(coord, c, cgfs)
        self.assertEqual(grad.shape[0], 6*6*6)
        self.assertEqual(grad.shape[1], 3)        
        res = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2_grad.npy'))
        np.testing.assert_almost_equal(grad, res, 4)

if __name__ == '__main__':
    unittest.main()



from .molecule import Molecule
from .pyqint import PyQInt
from .cgf import CGF
from .gto import GTO
from .hf import HF
from .foster_boys import FosterBoys
from .cohp import COHP
from .molecule_builder import MoleculeBuilder
from .geometry_optimization import GeometryOptimization
from .blenderrender import BlenderRender

from ._version import __version__



import unittest
from pyqint import GTO
import numpy as np

class testGTO(unittest.TestCase):

    def testNormalizationGTO(self):
        """
        Test getting amplitude from CGF
        """

        gto_h = GTO(1.0, (0,0,0), 0.4166, 0, 0, 0)
        norm = gto_h.get_norm()
        np.testing.assert_almost_equal(norm, 0.36957240951430304, 4)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import Molecule, CGF, HF, MoleculeBuilder
import numpy as np

class TestCustomBasisSet(unittest.TestCase):
    
    def test_custom_basis_set_h2(self):
        mol = Molecule()
        mol.add_atom('H', 0.0000, 0.0000, 0.3561150187, unit='angstrom')
        mol.add_atom('H', 0.0000, 0.0000, -0.3561150187, unit='angstrom')        
        nuclei = mol.get_nuclei()

        cgfs = []
        for n in nuclei:
            _cgf = CGF(n[0])

            _cgf.add_gto(0.154329, 3.425251, 0, 0, 0)
            _cgf.add_gto(0.535328, 0.623914, 0, 0, 0)
            _cgf.add_gto(0.444635, 0.168855, 0, 0, 0)

            cgfs.append(_cgf)

        res = HF().rhf(mol, basis=cgfs)
        np.testing.assert_almost_equal(res['energy'], -1.1175059, 5)

    def test_custom_basis_set_co(self):
        mol = MoleculeBuilder().from_name('CO')
        cgfs, nuclei = mol.build_basis('sto3g')

        res = HF().rhf(mol, basis=cgfs)
        np.testing.assert_almost_equal(res['energy'], -111.2192571, 4)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import PyQInt, GTO, Molecule
import numpy as np

class TestRepulsion(unittest.TestCase):

    def test_gto_repulsion(self):
        """
        Test two-electron integrals for primitive GTOs

        (ij|kl) = <gto_i gto_j | r_ij | gto_k gto_l>
        """

        # construct integrator object
        integrator = PyQInt()

        # test GTO
        gto1 = GTO(0.154329, [0.0, 0.0, 0.0], 3.425251, 0, 0, 0)
        gto2 = GTO(0.535328, [0.0, 0.0, 0.0], 0.623914, 0, 0, 0)
        gto3 = GTO(0.444635, [0.0, 0.0, 0.0], 0.168855, 0, 0, 0)
        repulsion = integrator.repulsion_gto(gto1, gto1, gto1, gto1)
        result = 0.20141123130697272
        np.testing.assert_almost_equal(repulsion, result, 4)

    def test_cgf_repulsion(self):
        """
        Test two-electron integrals for contracted Gaussians

        (ij|kl) = <cgf_i cgf_j | r_ij | cgf_k cgf_l>
        """

        # construct integrator object
        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.0, 0.0, 1.4)
        cgfs, nuclei = mol.build_basis('sto3g')

        T1111 = integrator.repulsion(cgfs[0], cgfs[0], cgfs[0], cgfs[0])
        T1122 = integrator.repulsion(cgfs[0], cgfs[0], cgfs[1], cgfs[1])
        T1112 = integrator.repulsion(cgfs[0], cgfs[0], cgfs[0], cgfs[1])
        T2121 = integrator.repulsion(cgfs[1], cgfs[0], cgfs[1], cgfs[0])
        T1222 = integrator.repulsion(cgfs[0], cgfs[1], cgfs[1], cgfs[1])
        T2211 = integrator.repulsion(cgfs[1], cgfs[1], cgfs[0], cgfs[0])

        np.testing.assert_almost_equal(T1111, 0.7746056914329529, 4)
        np.testing.assert_almost_equal(T1122, 0.5696758031845093, 4)
        np.testing.assert_almost_equal(T1112, 0.4441076656879812, 4)
        np.testing.assert_almost_equal(T2121, 0.2970285713672638, 4)

        # test similarity between two-electron integrals
        np.testing.assert_almost_equal(T1222, T1112, 4)
        np.testing.assert_almost_equal(T1122, T2211, 4)

    def test_two_electron_indices(self):
        """
        Test unique two-electron indices
        """
        integrator = PyQInt()

        np.testing.assert_almost_equal(integrator.teindex(1,1,2,1), integrator.teindex(1,1,1,2), 4)
        np.testing.assert_almost_equal(integrator.teindex(1,1,2,1), integrator.teindex(2,1,1,1), 4)
        np.testing.assert_almost_equal(integrator.teindex(1,2,1,1), integrator.teindex(2,1,1,1), 4)
        np.testing.assert_almost_equal(integrator.teindex(1,1,1,2), integrator.teindex(1,1,2,1), 4)
        self.assertNotEqual(integrator.teindex(1,1,1,1), integrator.teindex(1,1,2,1))
        self.assertNotEqual(integrator.teindex(1,1,2,1), integrator.teindex(1,1,2,2))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from . import HF, Molecule
import numpy as np
import scipy.optimize

class GeometryOptimization:
    """
    Class to perform geometry optimizaton using the Conjugate Gradient algorithm
    as implemented in the scipy library
    """
    def __init__(self, verbose=False):
        self.cinit = None
        self.P = None
        self.orbe = None
        self.verbose = verbose
        self.iter = 0
        self.energies_history = []
        self.forces_history = []
        self.coordinates_history = []
        self.coord = None

    def run(self, mol, basis, gtol=1e-5):
        """
        Perform geometry optimization
        """
        x0 = self.__unpack(mol) # unpack coordinates from molecule class
        self.iter = 0

        # reset arrays
        self.energies_history = []
        self.forces_history = []
        self.coordinates_history = []

        if self.verbose:
            self.__print_break(newline=False)
            print("START GEOMETRY OPTIMIZATION")
            print("USING CONJUGATE GRADIENT PROCEDURE")
            self.__print_break(newline=True)

        res_opt = scipy.optimize.minimize(self.energy, x0, args=(mol, basis), method='CG',
                                          jac=self.jacobian, options={'gtol':gtol})

        res = {
            'opt': res_opt,
            'energies': self.energies_history,
            'forces': self.forces_history,
            'coordinates': self.coordinates_history,
            'data': self.last_energy_run
        }

        return res

    def energy(self, x, mol, basis):
        self.iter += 1

        mol = self.__pack(mol, x) # pack positions into new molecule class

        if self.verbose:
            self.__print_break(newline=False)
            print('  START GEOMETRY OPTIMIZATION STEP %03i' % self.iter)
            self.__print_break()
            self.__print_positions(mol)
            print() # newline

        res = HF().rhf(mol, basis, orbc_init=self.cinit, calc_forces=True)

        if self.verbose:
            self.__print_energies(res)
            print() # print newline

        # cache matrices
        self.cinit = res['orbc']
        self.P = res['density']
        self.orbe = res['orbe']
        self.forces = res['forces']
        self.coord = x
        self.last_energy_run = res

        # append history
        self.forces_history.append(self.forces)
        self.coordinates_history.append(self.coord.reshape(len(mol.get_atoms()),3))
        self.energies_history.append(res['energies'][-1])

        if self.verbose:
            self.__print_forces(mol, self.forces)
            print() # print newline
            self.__print_break(newline=False)
            print('  END GEOMETRY OPTIMIZATION STEP %03i' % self.iter)
            self.__print_break(newline=True)

        return res['energies'][-1]

    def jacobian(self, x, mol, basis):
        """
        Calculate the forces. Note that the forces are typically already
        calculated in the energy step so here only a simple check is done
        to see if the coordinates match. If so, the forces are simply returned.
        """
        # check if forces are already calculated
        if self.coord is not None and np.max(x - self.coord) < 1e-5:
            return self.forces.flatten()
        else:
            # no forces have yet been calculated, indicative that no energy run has
            # yet been done
            res = HF().rhf(mol, basis, orbc_init=self.cinit, calc_forces=True)
            self.cinit = res['orbc']
            self.P = res['density']
            self.orbe = res['orbe']
            self.forces = res['forces']
            self.coord = x
            return res['forces'].flatten()

    def __unpack(self, mol):
        """
        Unpack coordinates from molecule class
        """
        coords = []
        for atom in mol.get_atoms():
            coords.append(atom[1])

        return np.array(coords).flatten()

    def __pack(self, mol, coords):
        """
        Pack coordinates into new molecule class
        """

        newmol = Molecule(mol.name)
        newmol.set_charge(mol.get_charge())
        coords = coords.reshape((len(mol.get_atoms()), 3))
        for i,atom in enumerate(mol.get_atoms()):
            newmol.add_atom(mol.get_atoms()[i][0], coords[i][0], coords[i][1], coords[i][2])

        return newmol

    def __print_positions(self, mol):
        """
        Print atomic positions in nice formatting
        """
        print('-------------')
        print('  POSITIONS  ')
        print('-------------')
        for atom in mol.get_atoms():
            print('  %2s %12.8f %12.8f %12.8f' % (atom[0],
                                                  atom[1][0],
                                                  atom[1][1],
                                                  atom[1][2]))

    def __print_forces(self, mol, forces):
        """
        Print forces using nice formatting
        """
        print('----------')
        print('  FORCES  ')
        print('----------')

        for atom,force in zip(mol.get_atoms(), forces):
            print('  %2s %12.4e %12.4e %12.4e' % (atom[0],
                                                  force[0],
                                                  force[1],
                                                  force[2]))

    def __print_energies(self, res):
        """
        Print the energy terms in each iteration
        """
        print('------------')
        print('  ENERGIES  ')
        print('------------')

        print('  Kinetic:                     %12.8f' % res['ekin'])
        print('  Nuclear:                     %12.8f' % res['enuc'])
        print('  Electron-electron repulsion: %12.8f' % res['erep'])
        print('  Exchange:                    %12.8f' % res['ex'])
        print('  Nuclear repulsion:           %12.8f' % res['enucrep'])
        print('  TOTAL:                       %12.8f' % res['energies'][-1])

    def __print_break(self, newline=True):
        print('=' * 80)
        if newline:
            print()
